#include <string.h>
#include <stdio.h>
#include "nsapp.h"
#include "hw.h"
#include "board.h"
#include "map.h"
#include "alg.h"

static void board_show_status(board_t *board, int run, int vld)
{
	char str[17];
	int i;
	memset(str, 0, 16);
	str[0] = board->name[7];
	if(board->alg[0])
		strcpy(str+2, board->alg[0]->sym);
	else
		strcpy(str+2, "--");
	if(vld) {
		sprintf(str+5, "%d/%d %c", board->len, board->pfx_len, run?'R':'-');
		if(board->percent >= 100 || !run)
			strcpy(str+13, "ALL");
		else
			sprintf(str+13, "%d%%", board->percent);
	} else {
		sprintf(str+5, "--- -");
		strcpy(str+13, "OFF");
	}
	for(i=0; i<16; i++)
		if(!str[i])
			str[i] = 32;
	str[16]=0;
	hw_led_line(board->hw, str);
}

static void board_rstep(unsigned char *buf, int n, int len);
static void board_fstep(unsigned char *buf, int n, int len);

static int board_process_result(unsigned chip, wu_t *wu, unsigned char *res)
{
	unsigned char str[MAX_LEN+1];
	unsigned buf[16], hash[MAX_HASH];
	int i, f;

	memset(str, 0, MAX_LEN+1);
	for(i=0; i<wu->task->len-wu->task->wu_size; i++)
		str[i] = wu->prefix[i];
	for(i=0; i<8; i++)
		if(wu->task->len>i)
			str[wu->task->len-1-i] ^= hamilton[res[i]];

	memset(buf, 0, 64);
	for(i=0; i<wu->task->len; i++)
		wu->task->alg->setbyte(i, str[i], buf);
	wu->task->alg->pad(wu->task->len, buf);

	wu->task->alg->hash(buf, hash);

	f=-1;
	for(i=0; i<wu->task->nhash; i++)
		if(!memcmp(hash, wu->task->hashes+(i*wu->task->hashlen), sizeof(unsigned)*wu->task->hashlen)) {
			f=i;
			break;
		}
	if(f==-1)
		return 1;

	for(i=0; i<wu->task->len; i++)
		if(str[i]<32 || str[i]>127)
			str[i] = '.';
	switch(wu->task->hashlen) {
	case 1: message(MSG_INFO, "%s.%c%d: %s | %08X", wu->board->name, 'A'+chip/5, chip%5, str, hash[0]); break;
	case 2: message(MSG_INFO, "%s.%c%d: %s | %08X %08X", wu->board->name, 'A'+chip/5, chip%5, str, hash[0], hash[1]); break;
	case 3: message(MSG_INFO, "%s.%c%d: %s | %08X %08X %08X", wu->board->name, 'A'+chip/5, chip%5, str, hash[0], hash[1], hash[2]); break;
	case 4: message(MSG_INFO, "%s.%c%d: %s | %08X %08X %08X %08X", wu->board->name, 'A'+chip/5, chip%5, str, hash[0], hash[1], hash[2], hash[3]); break;
	case 5: message(MSG_INFO, "%s.%c%d: %s | %08X %08X %08X %08X %08X", wu->board->name, 'A'+chip/5, chip%5, str, hash[0], hash[1], hash[2], hash[3], hash[4]); break;
	}

	return 0;
}

static int board_process_irq(board_t *board)
{
	unsigned char res[8], res2[8];
	unsigned chip, *val = board->irq_buf;
	int i, wu_done = 0, last;

	for(i=0; i<8; i++) {
		chip = (val[i]>>18)&15;
		if(chip == 15)
			retval_message(1, MSG_WARN, "Read timeout on chip %s.%c%d", board->name, 'A'+(val[i]&15)/5, (val[i]&15)%5);
		res[i] = (val[i]>>6)&63;
		res2[i] = (val[i]>>12)&63;
	}

	if(!board->wu[chip])
		retval_message(0, MSG_DBG, "%s.%c%d: WU finished (buffer skid)", board->name, 'A'+chip/5, chip%5);

	board_rstep(res, board->alg[chip]->latency, 8);
	board_rstep(res2, board->alg[chip]->latency, 8);

	for(i=board->wu[chip]->task->wu_size; i<8; i++)
		if(res[i])
			wu_done = 1;
	if(!wu_done)
		board_process_result(chip, board->wu[chip], res);
	while(1) {
		last = 1;
		for(i=0; i<8; i++)
			if(res[i] != res2[i])
				last = 0;
		if(last || wu_done)
			break;

		board_fstep(res, 1, 8);

		for(i=board->wu[chip]->task->wu_size; i<8; i++)
			if(res[i])
				wu_done = 1;
		if(!wu_done)
			board_process_result(chip, board->wu[chip], res);
	}

	if(wu_done) {
		message(MSG_DBG, "%s.%c%d: WU finished", board->name, 'A'+chip/5, chip%5);
		board->irq_mask &= ~(1<<chip);
		hw_irq_mask(board->hw, board->irq_mask);
		hw_chip_mask(board->hw, 1<<chip);
		hw_write_ctrl(board->hw, 0x0001);	// stop & reset
		hw_read_reg(board->hw, 8, 1);		// read status
		last = wu_push_done(board->wu[chip]);
		if(last != board->percent) {
			board->percent = last;
			board_show_status(board, 1, 1);
		}
		board->wu[chip] = NULL;
	}

	return 0;
}

static int board_load_map(board_t *board, map_t *map)
{
	unsigned lev, adr, val, base, bits;
	if(hw_chip_mask(board->hw, 0x7FFF))
		goto fail;
	for(lev=0; lev<map->nlev; lev++) {
		message(MSG_DBG, "%s: Loading map level %d", board->name, lev);
		base = lev<<10;
		bits = map->bits[lev];
		for(adr=0; adr<map->sizes[lev]; adr++) {
			val = map->maps[lev][adr];
			if(bits>18) {
				if(hw_write_reg(board->hw, base+adr*2, val))
					goto fail;
				if(hw_write_reg(board->hw, base+adr*2+1, val>>(bits-18)))
					goto fail;
			} else
				if(hw_write_reg(board->hw, base+adr, val))
					goto fail;
		}
	}
	board->map = map;
	retval_message(0, MSG_INFO, "%s: Map loaded", board->name);
fail:
	retval_message(1, MSG_FAIL, "%s: Loading map failed", board->name);
}

static unsigned char getbyte(int idx, unsigned *src)
{
	int shift = (idx%4)*8;
	return (src[idx/4]>>shift)&0xFF;
}

static int board_load_wu(board_t *board, int chip, wu_t *wu)
{
	int i;
	unsigned buf[16];
	unsigned char b;
	message(MSG_DBG, "%s: accepted WU '%s' on chip %d", board->name, wu->prefix, chip);

	wu->board = board;
	wu->chip = chip;
	board->wu[chip] = wu;
	board->ready[chip] = 0;
	if(board->map != wu->task->map)
		board_load_map(board, wu->task->map);

	if(board->len != wu->task->len || board->pfx_len != wu->task->pfx_len) {
		board->len = wu->task->len;
		board->pfx_len = wu->task->pfx_len;
		board_show_status(board, 1, 1);
	}

	hw_chip_mask(board->hw, 1<<chip);
	hw_write_ctrl(board->hw, 0x0001);				// reset
	hw_write_ctrl(board->hw, 0x0000|((wu->task->len-1)<<6));	// stop, set length
	for(i=0; i<wu->task->len-wu->task->wu_size; i++)
		hw_xor_byte(board->hw, i, wu->prefix[i]);

	// create and write padding
	memset(buf, 0, 64);
	board->alg[chip]->pad(wu->task->len, buf);
	for(i=0; i<64; i++) {
		b = getbyte(i, buf);
		if(b)
			hw_xor_byte(board->hw, i, b);
	}
	hw_write_ctrl(board->hw, 0x0002|((wu->task->len-1)<<6));	// run

	board->irq_mask |= 1<<chip;
	hw_irq_mask(board->hw, board->irq_mask);
	return 0;
}

void *board_thread(void *ptr)
{
	int i, done, busy, chip;
	unsigned res;
	wu_t *new_wu;
	board_t *board = (board_t *)ptr;

	board->hw = hw_open_board(board->name);
	if(!board->hw)
		return NULL;

	hw_led_line(board->hw, board->name);
	board->percent = board->len = 0;

	hw_chip_mask(board->hw, 0x7FFF);
	hw_read_reg(board->hw, 8, 1);
	for(i=0;i<15;i++) {
		hw_read_word(board->hw, &res, 1000000);
		chip = (res>>18)&15;
		board->ready[chip]=1;
		board->alg[chip]=alg_find_id((res>>12)&0x3F);
		if(board->alg[chip]) {
		    message(MSG_DBG, "%s.%c%d status: %05X (algorithm: %s)", board->name, 'A'+chip/5, chip%5, res&0x3FFFF, board->alg[chip]->name);
		} else
		    message(MSG_DBG, "%s.%c%d status: %05X (unknown algorithm)", board->name, 'A'+chip/5, chip%5, res&0x3FFFF);
	}
	board_show_status(board, 0, 0);

	do {
		i = hw_read_word(board->hw, &res, 1000000);
		if(i == 1) {
			// failure! kill all work units and force map reload D:
			done = 1;
			continue;
		}
		if(i == 0) {
			if(res&0x800000) {
				board->irq_buf[board->irq_ptr++] = res;
				if(board->irq_ptr == 8) {
					board_process_irq(board);
					board->irq_ptr = 0;
				}
			} else {
				message(MSG_DBG, "%s.%c%d status: %05X", board->name, 'A'+((res>>18)&15)/5, ((res>>18)&15)%5, res&0x3FFFF);
				board->ready[(res>>18)&15]=1;
			}
		}

		busy = 0;
		for(i=0; i<15; i++)
			if(board->wu[i] || !board->ready[i])
				busy = 1;
		done = !busy;
		for(i=0; i<15; i++)
			if(!board->wu[i] && board->ready[i]) {
				new_wu = wu_pop_todo();
				if(!new_wu) {
					wu_unlock();
					break;
				}
				done = 0;
				if((busy && new_wu->task->map != board->map) || (new_wu->task->alg != board->alg[i])) {
					wu_push_todo(new_wu);
					break;
				}
				wu_unlock();
				board_load_wu(board, i, new_wu);
			}
	} while(!done);

	if(board->len)
		board_show_status(board, 0, 1);
	usleep(100000);

	hw_close_board(board->hw);
	return NULL;
}

static void board_rstep(unsigned char *buf, int n, int len)
{
    int r = buf[0], i;
    unsigned char xv = 0;
    for(i=1;i<len;i++)
	xv ^= buf[i];
    if(xv&1)
	r += n;
    else
	r -= n;
    while(1) {
        if(r >= 64) {
	    r = 127-r;
	    board_rstep(buf+1, 1, len-1);
	} else if(r < 0) {
	    r = -1-r;
	    board_rstep(buf+1, 1, len-1);
	} else
	    break;
    }
    buf[0] = r;
}

static void board_fstep(unsigned char *buf, int n, int len)
{
    int r = buf[0], i;
    unsigned char xv = 0;
    for(i=1;i<len;i++)
	xv ^= buf[i];
    if(xv&1)
	r -= n;
    else
	r += n;
    while(1) {
        if(r >= 64) {
	    r = 127-r;
	    board_fstep(buf+1, 1, len-1);
	} else if(r < 0) {
	    r = -1-r;
	    board_fstep(buf+1, 1, len-1);
	} else
	    break;
    }
    buf[0] = r;
}
