#include <malloc.h>
#include "nsapp.h"
#include "map.h"

#define NLEV 5

//           ptr  cmp  flags bits  rams   size
// array 0:  10    1     2    13    1x  1024x18
// array 1:  10    2     2    14    1x  1024x18
// array 2:  10    2     2    14    1x  1024x18
// array 3:   9---17-----2    18    1x  1024x18
// array 4:       28     2    30    1x   512x32
static unsigned bits[NLEV]={10,1,2,2,17};
static unsigned l2size[NLEV]={10,10,10,10,9};
static unsigned size[NLEV]={1024,1024,1024,1024,512};

struct mapctx {
	unsigned *nmap[NLEV];
	unsigned *vmap[NLEV];
	unsigned alloc[NLEV+1];
	unsigned *pmap[NLEV];
	unsigned *cmap[NLEV];
};

#define LINEBITS 3

static void prepare(struct mapctx *c)
{
    int i;
    for(i=0;i<NLEV;i++) {
	c->nmap[i]=calloc(sizeof(unsigned), size[i]);
	c->pmap[i]=calloc(sizeof(unsigned), size[i]);
	c->cmap[i]=calloc(sizeof(unsigned), size[i]);
	c->vmap[i]=calloc(sizeof(unsigned), size[i]);
	c->alloc[i]=0;
    }
    c->alloc[NLEV]=0;
}

static void free_all(struct mapctx *c)
{
    int i;
    for(i=0;i<NLEV;i++) {
	free(c->nmap[i]);
	free(c->pmap[i]);
	free(c->cmap[i]);
	free(c->vmap[i]);
    }
}

static int map_level(int lev, struct mapctx *c, int nnid, unsigned *nids)
{
    int i, j;
    int cstat=0, ustat=0;
    unsigned t, v=0, pv=0, pb=0, pc=0, px=0;
    for(i=0;i<nnid;i++) {
	pv=pb=pc=px=0;
	t=nids[i];
	for(j=0;j<=lev && pb!=-1;j++) {
	    if(j==NLEV-1)
		v=t&((1<<LINEBITS)-1);
	    else
		v=t&((1<<bits[j])-1);
	    if(pc) {
		if(px!=v)
		    retval_message(0, MSG_FAIL, "Internal error at level %d (lev=%d, px=%d, v=%d)", lev, j, px, v);
		pv=pb;
	    } else
		pv=pb+v;
	    if(j==NLEV-1)
		t>>=LINEBITS;
	    else
		t>>=bits[j];
	    pb=c->pmap[j][pv];
	    pc=c->cmap[j][pv];
	    px=c->vmap[j][pv];
	}
	if(pb!=-1) {
	    if(lev==NLEV-1) {
		c->vmap[lev][pv]|=(t|(1<<(bits[lev]-LINEBITS)))<<(c->nmap[lev][pv]*(1+bits[lev]-LINEBITS));
		c->nmap[lev][pv]++;
		if(c->nmap[lev][pv]>2)
		    retval_message(0, MSG_FAIL, "Level %d: exceeded 2 ways", lev);
	    } else {
		v=t&((1<<bits[lev+1])-1);
		if(px!=v || !c->nmap[lev][pv])
		    c->nmap[lev][pv]++;
		c->vmap[lev][pv]=v;
	    }
	}
    }
    if(lev!=NLEV-1) {
	for(i=0;i<size[lev];i++)
	    if(c->nmap[lev][i]) {
		if(c->nmap[lev][i]==1) {
		    c->cmap[lev][i]=1;
		    if(lev==NLEV-2)
			c->pmap[lev][i]=-1;
		    else {
			c->pmap[lev][i]=c->alloc[lev+1];
			c->alloc[lev+1]++;
		    }
		    cstat++;
		} else {
		    c->cmap[lev][i]=0;
		    c->vmap[lev][i]=0;
		    c->pmap[lev][i]=c->alloc[lev+1];
		    if(lev==NLEV-2) {
			if(c->nmap[lev][i]>(2<<LINEBITS))
			    retval_message(0, MSG_FAIL, "Level %d: exceeded %d per line", lev+1, 2<<LINEBITS);
			c->alloc[lev+1]+=1<<LINEBITS;
		    } else
			c->alloc[lev+1]+=1<<bits[lev+1];
		    ustat++;
		}
		if(lev+1<NLEV && c->alloc[lev+1]>size[lev+1])
		    retval_message(0, MSG_FAIL, "Level %d: exceeded %d", lev+1, size[lev+1]);
	    }
	if(lev+1<NLEV) {
	    if(lev+2<NLEV)
		message(MSG_INFO, "Level %d: %d/%d (%d uncompressed, %d compressed)", lev+1, c->alloc[lev+1], size[lev+1], ustat, cstat);
	    else
		message(MSG_INFO, "Level %d: %d/%d (%d uncompressed, %d collapsed)", lev+1, c->alloc[lev+1], size[lev+1], ustat, cstat);
	}
    }
    return 1;
}

static int map_all(struct mapctx *c, int nnid, unsigned *nids)
{
    int i;
    message(MSG_INFO, "Level 0: %d", size[0]);
    for(i=0;i<NLEV;i++)
	if(!map_level(i, c, nnid, nids))
	    return 0;
    return 1;
}

static void dump_all(struct mapctx *c, map_t *map)
{
    int lev, i;
    unsigned code, nbit;
    for(lev=0;lev<NLEV-1;lev++) {
	nbit=l2size[lev+1]+bits[lev+1]+2;
	if(lev==NLEV-2)
	    nbit-=l2size[lev+1]+1;
	map->sizes[lev] = size[lev];
	map->bits[lev] = nbit;
	for(i=0;i<size[lev];i++) {
	    code=c->pmap[lev][i];
	    if(code==-1) {
		if(!c->cmap[lev][i])
		    return_message(MSG_FAIL, "Internal error: uncompressed pointer collapsed");
		code=0;
	    }
	    if(lev==NLEV-2) {
		// we alias pointer and compare for the next-to-last level, since they can't alias
		// also we alias valid and compare, since if compare is true then compress is true anyway
		code|=c->vmap[lev][i];
		if(c->nmap[lev][i] && !c->cmap[lev][i])
		    code|=1<<(bits[lev+1]-1);
		code|=c->cmap[lev][i]<<bits[lev+1];
	    } else {
		code|=c->vmap[lev][i]<<l2size[lev+1];
		if(c->nmap[lev][i])
		    code|=1<<(l2size[lev+1]+bits[lev+1]+1);
		code|=c->cmap[lev][i]<<(l2size[lev+1]+bits[lev+1]);
	    }
	    map->maps[lev][i] = code;
	}
    }
    lev=NLEV-1;
    nbit=2*(bits[lev]-LINEBITS+1);
    map->sizes[lev] = size[lev];
    map->bits[lev] = nbit;
    for(i=0;i<size[lev];i++)
	map->maps[lev][i] = c->vmap[lev][i];
}

map_t *map_compile(int nnid, unsigned *nids)
{
    int i;
    struct mapctx ctx;
    map_t *map = NULL;
    message(MSG_INFO, "Mapping %d NIDs", nnid);
    prepare(&ctx);
    if(map_all(&ctx, nnid, nids)) {
    	map = calloc(1, sizeof(map_t));
	map->nlev = NLEV;
	map->sizes = calloc(NLEV, sizeof(int));
	map->bits = calloc(NLEV, sizeof(int));
	map->maps = calloc(NLEV, sizeof(unsigned *));
	for(i=0; i<NLEV; i++)
		map->maps[i] = calloc(size[i], sizeof(unsigned));
	dump_all(&ctx, map);
    }
    free_all(&ctx);
    return map;
}

void map_free(map_t *map)
{
	int i;
	for(i=0;i<map->nlev;i++)
		free(map->maps[i]);
	free(map->sizes);
	free(map->bits);
	free(map);
}
